# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("winbio.dll")
prototypes = \
    {
        #
        'WinBioEnumServiceProviders': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"BiometricFactor": SimTypeInt(signed=False, label="UInt32"), "BspId": SimTypeBottom(label="Guid"), "Description": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Vendor": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Version": SimStruct({"MajorVersion": SimTypeInt(signed=False, label="UInt32"), "MinorVersion": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_VERSION", pack=False, align=None)}, name="WINBIO_BSP_SCHEMA", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Factor", "BspSchemaArray", "BspCount"]),
        #
        'WinBioEnumBiometricUnits': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"UnitId": SimTypeInt(signed=False, label="UInt32"), "PoolType": SimTypeInt(signed=False, label="UInt32"), "BiometricFactor": SimTypeInt(signed=False, label="UInt32"), "SensorSubType": SimTypeInt(signed=False, label="UInt32"), "Capabilities": SimTypeInt(signed=False, label="UInt32"), "DeviceInstanceId": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Description": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Manufacturer": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Model": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "SerialNumber": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "FirmwareVersion": SimStruct({"MajorVersion": SimTypeInt(signed=False, label="UInt32"), "MinorVersion": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_VERSION", pack=False, align=None)}, name="WINBIO_UNIT_SCHEMA", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Factor", "UnitSchemaArray", "UnitCount"]),
        #
        'WinBioEnumDatabases': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimStruct({"BiometricFactor": SimTypeInt(signed=False, label="UInt32"), "DatabaseId": SimTypeBottom(label="Guid"), "DataFormat": SimTypeBottom(label="Guid"), "Attributes": SimTypeInt(signed=False, label="UInt32"), "FilePath": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "ConnectionString": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256)}, name="WINBIO_STORAGE_SCHEMA", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Factor", "StorageSchemaArray", "StorageCount"]),
        #
        'WinBioAsyncOpenFramework': SimTypeFunction([SimTypeInt(signed=False, label="WINBIO_ASYNC_NOTIFICATION_METHOD"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimStruct({"SessionHandle": SimTypeInt(signed=False, label="UInt32"), "Operation": SimTypeInt(signed=False, label="UInt32"), "SequenceNumber": SimTypeLongLong(signed=False, label="UInt64"), "TimeStamp": SimTypeLongLong(signed=True, label="Int64"), "ApiStatus": SimTypeBottom(label="HRESULT"), "UnitId": SimTypeInt(signed=False, label="UInt32"), "UserData": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "Parameters": SimUnion({"Verify": SimStruct({"Match": SimTypeBottom(label="BOOLEAN"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_Verify_e__Struct", pack=False, align=None), "Identify": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_Identify_e__Struct", pack=False, align=None), "EnrollBegin": SimStruct({"SubFactor": SimTypeChar(label="Byte")}, name="_EnrollBegin_e__Struct", pack=False, align=None), "EnrollCapture": SimStruct({"RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_EnrollCapture_e__Struct", pack=False, align=None), "EnrollCommit": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "IsNewTemplate": SimTypeBottom(label="BOOLEAN")}, name="_EnrollCommit_e__Struct", pack=False, align=None), "EnumEnrollments": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactorCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "SubFactorArray": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="_EnumEnrollments_e__Struct", pack=False, align=None), "CaptureSample": SimStruct({"Sample": SimTypePointer(SimStruct({"HeaderBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "StandardDataBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "VendorDataBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "SignatureBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None)}, name="WINBIO_BIR", pack=False, align=None), offset=0), "SampleSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_CaptureSample_e__Struct", pack=False, align=None), "DeleteTemplate": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte")}, name="_DeleteTemplate_e__Struct", pack=False, align=None), "GetProperty": SimStruct({"PropertyType": SimTypeInt(signed=False, label="UInt32"), "PropertyId": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "PropertyBufferSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "PropertyBuffer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="_GetProperty_e__Struct", pack=False, align=None), "SetProperty": SimStruct({"PropertyType": SimTypeInt(signed=False, label="UInt32"), "PropertyId": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "PropertyBufferSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "PropertyBuffer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="_SetProperty_e__Struct", pack=False, align=None), "GetEvent": SimStruct({"Event": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Parameters": SimUnion({"Unclaimed": SimStruct({"UnitId": SimTypeInt(signed=False, label="UInt32"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_Unclaimed_e__Struct", pack=False, align=None), "UnclaimedIdentify": SimStruct({"UnitId": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_UnclaimedIdentify_e__Struct", pack=False, align=None), "Error": SimStruct({"ErrorCode": SimTypeBottom(label="HRESULT")}, name="_Error_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="WINBIO_EVENT", pack=False, align=None)}, name="_GetEvent_e__Struct", pack=False, align=None), "ControlUnit": SimStruct({"Component": SimTypeInt(signed=False, label="WINBIO_COMPONENT"), "ControlCode": SimTypeInt(signed=False, label="UInt32"), "OperationStatus": SimTypeInt(signed=False, label="UInt32"), "SendBuffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "SendBufferSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "ReceiveBuffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ReceiveBufferSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "ReceiveDataSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="_ControlUnit_e__Struct", pack=False, align=None), "EnumServiceProviders": SimStruct({"BspCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "BspSchemaArray": SimTypePointer(SimStruct({"BiometricFactor": SimTypeInt(signed=False, label="UInt32"), "BspId": SimTypeBottom(label="Guid"), "Description": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Vendor": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Version": SimStruct({"MajorVersion": SimTypeInt(signed=False, label="UInt32"), "MinorVersion": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_VERSION", pack=False, align=None)}, name="WINBIO_BSP_SCHEMA", pack=False, align=None), offset=0)}, name="_EnumServiceProviders_e__Struct", pack=False, align=None), "EnumBiometricUnits": SimStruct({"UnitCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "UnitSchemaArray": SimTypePointer(SimStruct({"UnitId": SimTypeInt(signed=False, label="UInt32"), "PoolType": SimTypeInt(signed=False, label="UInt32"), "BiometricFactor": SimTypeInt(signed=False, label="UInt32"), "SensorSubType": SimTypeInt(signed=False, label="UInt32"), "Capabilities": SimTypeInt(signed=False, label="UInt32"), "DeviceInstanceId": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Description": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Manufacturer": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Model": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "SerialNumber": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "FirmwareVersion": SimStruct({"MajorVersion": SimTypeInt(signed=False, label="UInt32"), "MinorVersion": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_VERSION", pack=False, align=None)}, name="WINBIO_UNIT_SCHEMA", pack=False, align=None), offset=0)}, name="_EnumBiometricUnits_e__Struct", pack=False, align=None), "EnumDatabases": SimStruct({"StorageCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "StorageSchemaArray": SimTypePointer(SimStruct({"BiometricFactor": SimTypeInt(signed=False, label="UInt32"), "DatabaseId": SimTypeBottom(label="Guid"), "DataFormat": SimTypeBottom(label="Guid"), "Attributes": SimTypeInt(signed=False, label="UInt32"), "FilePath": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "ConnectionString": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256)}, name="WINBIO_STORAGE_SCHEMA", pack=False, align=None), offset=0)}, name="_EnumDatabases_e__Struct", pack=False, align=None), "VerifyAndReleaseTicket": SimStruct({"Match": SimTypeBottom(label="BOOLEAN"), "RejectDetail": SimTypeInt(signed=False, label="UInt32"), "Ticket": SimTypeLongLong(signed=False, label="UInt64")}, name="_VerifyAndReleaseTicket_e__Struct", pack=False, align=None), "IdentifyAndReleaseTicket": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "RejectDetail": SimTypeInt(signed=False, label="UInt32"), "Ticket": SimTypeLongLong(signed=False, label="UInt64")}, name="_IdentifyAndReleaseTicket_e__Struct", pack=False, align=None), "EnrollSelect": SimStruct({"SelectorValue": SimTypeLongLong(signed=False, label="UInt64")}, name="_EnrollSelect_e__Struct", pack=False, align=None), "MonitorPresence": SimStruct({"ChangeType": SimTypeInt(signed=False, label="UInt32"), "PresenceCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "PresenceArray": SimTypePointer(SimStruct({"Factor": SimTypeInt(signed=False, label="UInt32"), "SubFactor": SimTypeChar(label="Byte"), "Status": SimTypeBottom(label="HRESULT"), "RejectDetail": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "TrackingId": SimTypeLongLong(signed=False, label="UInt64"), "Ticket": SimTypeLongLong(signed=False, label="UInt64"), "Properties": SimUnion({"FacialFeatures": SimStruct({"BoundingBox": SimTypeBottom(label="RECT"), "Distance": SimTypeInt(signed=True, label="Int32"), "OpaqueEngineData": SimStruct({"AdapterId": SimTypeBottom(label="Guid"), "Data": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 77)}, name="_OpaqueEngineData_e__Struct", pack=False, align=None)}, name="_FacialFeatures_e__Struct", pack=False, align=None), "Iris": SimStruct({"EyeBoundingBox_1": SimTypeBottom(label="RECT"), "EyeBoundingBox_2": SimTypeBottom(label="RECT"), "PupilCenter_1": SimTypeBottom(label="POINT"), "PupilCenter_2": SimTypeBottom(label="POINT"), "Distance": SimTypeInt(signed=True, label="Int32")}, name="_Iris_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "Authorization": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="_Authorization_e__Struct", pack=False, align=None)}, name="WINBIO_PRESENCE", pack=False, align=None), offset=0)}, name="_MonitorPresence_e__Struct", pack=False, align=None), "GetProtectionPolicy": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "Policy": SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "DatabaseId": SimTypeBottom(label="Guid"), "UserState": SimTypeLongLong(signed=False, label="UInt64"), "PolicySize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "Policy": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 128)}, name="WINBIO_PROTECTION_POLICY", pack=False, align=None)}, name="_GetProtectionPolicy_e__Struct", pack=False, align=None), "NotifyUnitStatusChange": SimStruct({"ExtendedStatus": SimStruct({"Availability": SimTypeInt(signed=False, label="UInt32"), "ReasonCode": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_EXTENDED_UNIT_STATUS", pack=False, align=None)}, name="_NotifyUnitStatusChange_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="WINBIO_ASYNC_RESULT", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["AsyncResult"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["NotificationMethod", "TargetWindow", "MessageCode", "CallbackRoutine", "UserData", "AsynchronousOpen", "FrameworkHandle"]),
        #
        'WinBioCloseFramework': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["FrameworkHandle"]),
        #
        'WinBioAsyncEnumServiceProviders': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["FrameworkHandle", "Factor"]),
        #
        'WinBioAsyncEnumBiometricUnits': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["FrameworkHandle", "Factor"]),
        #
        'WinBioAsyncEnumDatabases': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["FrameworkHandle", "Factor"]),
        #
        'WinBioAsyncMonitorFrameworkChanges': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["FrameworkHandle", "ChangeTypes"]),
        #
        'WinBioOpenSession': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="WINBIO_POOL"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Factor", "PoolType", "Flags", "UnitArray", "UnitCount", "DatabaseId", "SessionHandle"]),
        #
        'WinBioAsyncOpenSession': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="WINBIO_POOL"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeInt(signed=False, label="WINBIO_ASYNC_NOTIFICATION_METHOD"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimStruct({"SessionHandle": SimTypeInt(signed=False, label="UInt32"), "Operation": SimTypeInt(signed=False, label="UInt32"), "SequenceNumber": SimTypeLongLong(signed=False, label="UInt64"), "TimeStamp": SimTypeLongLong(signed=True, label="Int64"), "ApiStatus": SimTypeBottom(label="HRESULT"), "UnitId": SimTypeInt(signed=False, label="UInt32"), "UserData": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "Parameters": SimUnion({"Verify": SimStruct({"Match": SimTypeBottom(label="BOOLEAN"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_Verify_e__Struct", pack=False, align=None), "Identify": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_Identify_e__Struct", pack=False, align=None), "EnrollBegin": SimStruct({"SubFactor": SimTypeChar(label="Byte")}, name="_EnrollBegin_e__Struct", pack=False, align=None), "EnrollCapture": SimStruct({"RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_EnrollCapture_e__Struct", pack=False, align=None), "EnrollCommit": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "IsNewTemplate": SimTypeBottom(label="BOOLEAN")}, name="_EnrollCommit_e__Struct", pack=False, align=None), "EnumEnrollments": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactorCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "SubFactorArray": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="_EnumEnrollments_e__Struct", pack=False, align=None), "CaptureSample": SimStruct({"Sample": SimTypePointer(SimStruct({"HeaderBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "StandardDataBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "VendorDataBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "SignatureBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None)}, name="WINBIO_BIR", pack=False, align=None), offset=0), "SampleSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_CaptureSample_e__Struct", pack=False, align=None), "DeleteTemplate": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte")}, name="_DeleteTemplate_e__Struct", pack=False, align=None), "GetProperty": SimStruct({"PropertyType": SimTypeInt(signed=False, label="UInt32"), "PropertyId": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "PropertyBufferSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "PropertyBuffer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="_GetProperty_e__Struct", pack=False, align=None), "SetProperty": SimStruct({"PropertyType": SimTypeInt(signed=False, label="UInt32"), "PropertyId": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "PropertyBufferSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "PropertyBuffer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="_SetProperty_e__Struct", pack=False, align=None), "GetEvent": SimStruct({"Event": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Parameters": SimUnion({"Unclaimed": SimStruct({"UnitId": SimTypeInt(signed=False, label="UInt32"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_Unclaimed_e__Struct", pack=False, align=None), "UnclaimedIdentify": SimStruct({"UnitId": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_UnclaimedIdentify_e__Struct", pack=False, align=None), "Error": SimStruct({"ErrorCode": SimTypeBottom(label="HRESULT")}, name="_Error_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="WINBIO_EVENT", pack=False, align=None)}, name="_GetEvent_e__Struct", pack=False, align=None), "ControlUnit": SimStruct({"Component": SimTypeInt(signed=False, label="WINBIO_COMPONENT"), "ControlCode": SimTypeInt(signed=False, label="UInt32"), "OperationStatus": SimTypeInt(signed=False, label="UInt32"), "SendBuffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "SendBufferSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "ReceiveBuffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ReceiveBufferSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "ReceiveDataSize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="_ControlUnit_e__Struct", pack=False, align=None), "EnumServiceProviders": SimStruct({"BspCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "BspSchemaArray": SimTypePointer(SimStruct({"BiometricFactor": SimTypeInt(signed=False, label="UInt32"), "BspId": SimTypeBottom(label="Guid"), "Description": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Vendor": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Version": SimStruct({"MajorVersion": SimTypeInt(signed=False, label="UInt32"), "MinorVersion": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_VERSION", pack=False, align=None)}, name="WINBIO_BSP_SCHEMA", pack=False, align=None), offset=0)}, name="_EnumServiceProviders_e__Struct", pack=False, align=None), "EnumBiometricUnits": SimStruct({"UnitCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "UnitSchemaArray": SimTypePointer(SimStruct({"UnitId": SimTypeInt(signed=False, label="UInt32"), "PoolType": SimTypeInt(signed=False, label="UInt32"), "BiometricFactor": SimTypeInt(signed=False, label="UInt32"), "SensorSubType": SimTypeInt(signed=False, label="UInt32"), "Capabilities": SimTypeInt(signed=False, label="UInt32"), "DeviceInstanceId": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Description": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Manufacturer": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "Model": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "SerialNumber": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "FirmwareVersion": SimStruct({"MajorVersion": SimTypeInt(signed=False, label="UInt32"), "MinorVersion": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_VERSION", pack=False, align=None)}, name="WINBIO_UNIT_SCHEMA", pack=False, align=None), offset=0)}, name="_EnumBiometricUnits_e__Struct", pack=False, align=None), "EnumDatabases": SimStruct({"StorageCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "StorageSchemaArray": SimTypePointer(SimStruct({"BiometricFactor": SimTypeInt(signed=False, label="UInt32"), "DatabaseId": SimTypeBottom(label="Guid"), "DataFormat": SimTypeBottom(label="Guid"), "Attributes": SimTypeInt(signed=False, label="UInt32"), "FilePath": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256), "ConnectionString": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 256)}, name="WINBIO_STORAGE_SCHEMA", pack=False, align=None), offset=0)}, name="_EnumDatabases_e__Struct", pack=False, align=None), "VerifyAndReleaseTicket": SimStruct({"Match": SimTypeBottom(label="BOOLEAN"), "RejectDetail": SimTypeInt(signed=False, label="UInt32"), "Ticket": SimTypeLongLong(signed=False, label="UInt64")}, name="_VerifyAndReleaseTicket_e__Struct", pack=False, align=None), "IdentifyAndReleaseTicket": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "RejectDetail": SimTypeInt(signed=False, label="UInt32"), "Ticket": SimTypeLongLong(signed=False, label="UInt64")}, name="_IdentifyAndReleaseTicket_e__Struct", pack=False, align=None), "EnrollSelect": SimStruct({"SelectorValue": SimTypeLongLong(signed=False, label="UInt64")}, name="_EnrollSelect_e__Struct", pack=False, align=None), "MonitorPresence": SimStruct({"ChangeType": SimTypeInt(signed=False, label="UInt32"), "PresenceCount": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "PresenceArray": SimTypePointer(SimStruct({"Factor": SimTypeInt(signed=False, label="UInt32"), "SubFactor": SimTypeChar(label="Byte"), "Status": SimTypeBottom(label="HRESULT"), "RejectDetail": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "TrackingId": SimTypeLongLong(signed=False, label="UInt64"), "Ticket": SimTypeLongLong(signed=False, label="UInt64"), "Properties": SimUnion({"FacialFeatures": SimStruct({"BoundingBox": SimTypeBottom(label="RECT"), "Distance": SimTypeInt(signed=True, label="Int32"), "OpaqueEngineData": SimStruct({"AdapterId": SimTypeBottom(label="Guid"), "Data": SimTypeFixedSizeArray(SimTypeInt(signed=False, label="UInt32"), 77)}, name="_OpaqueEngineData_e__Struct", pack=False, align=None)}, name="_FacialFeatures_e__Struct", pack=False, align=None), "Iris": SimStruct({"EyeBoundingBox_1": SimTypeBottom(label="RECT"), "EyeBoundingBox_2": SimTypeBottom(label="RECT"), "PupilCenter_1": SimTypeBottom(label="POINT"), "PupilCenter_2": SimTypeBottom(label="POINT"), "Distance": SimTypeInt(signed=True, label="Int32")}, name="_Iris_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "Authorization": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="_Authorization_e__Struct", pack=False, align=None)}, name="WINBIO_PRESENCE", pack=False, align=None), offset=0)}, name="_MonitorPresence_e__Struct", pack=False, align=None), "GetProtectionPolicy": SimStruct({"Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "Policy": SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "DatabaseId": SimTypeBottom(label="Guid"), "UserState": SimTypeLongLong(signed=False, label="UInt64"), "PolicySize": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "Policy": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 128)}, name="WINBIO_PROTECTION_POLICY", pack=False, align=None)}, name="_GetProtectionPolicy_e__Struct", pack=False, align=None), "NotifyUnitStatusChange": SimStruct({"ExtendedStatus": SimStruct({"Availability": SimTypeInt(signed=False, label="UInt32"), "ReasonCode": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_EXTENDED_UNIT_STATUS", pack=False, align=None)}, name="_NotifyUnitStatusChange_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="WINBIO_ASYNC_RESULT", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["AsyncResult"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Factor", "PoolType", "Flags", "UnitArray", "UnitCount", "DatabaseId", "NotificationMethod", "TargetWindow", "MessageCode", "CallbackRoutine", "UserData", "AsynchronousOpen", "SessionHandle"]),
        #
        'WinBioCloseSession': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle"]),
        #
        'WinBioVerify': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "Identity", "SubFactor", "UnitId", "Match", "RejectDetail"]),
        #
        'WinBioVerifyWithCallback': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="HRESULT"), SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="BOOLEAN"), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["VerifyCallbackContext", "OperationStatus", "UnitId", "Match", "RejectDetail"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "Identity", "SubFactor", "VerifyCallback", "VerifyCallbackContext"]),
        #
        'WinBioIdentify': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "UnitId", "Identity", "SubFactor", "RejectDetail"]),
        #
        'WinBioIdentifyWithCallback': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="HRESULT"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypeChar(label="Byte"), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["IdentifyCallbackContext", "OperationStatus", "UnitId", "Identity", "SubFactor", "RejectDetail"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "IdentifyCallback", "IdentifyCallbackContext"]),
        #
        'WinBioWait': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle"]),
        #
        'WinBioCancel': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle"]),
        #
        'WinBioLocateSensor': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "UnitId"]),
        #
        'WinBioLocateSensorWithCallback': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="HRESULT"), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["LocateCallbackContext", "OperationStatus", "UnitId"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "LocateCallback", "LocateCallbackContext"]),
        #
        'WinBioEnrollBegin': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeChar(label="Byte"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "SubFactor", "UnitId"]),
        #
        'WinBioEnrollSelect': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeLongLong(signed=False, label="UInt64")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "SelectorValue"]),
        #
        'WinBioEnrollCapture': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "RejectDetail"]),
        #
        'WinBioEnrollCaptureWithCallback': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="HRESULT"), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["EnrollCallbackContext", "OperationStatus", "RejectDetail"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "EnrollCallback", "EnrollCallbackContext"]),
        #
        'WinBioEnrollCommit': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "Identity", "IsNewTemplate"]),
        #
        'WinBioEnrollDiscard': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle"]),
        #
        'WinBioEnumEnrollments': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "UnitId", "Identity", "SubFactorArray", "SubFactorCount"]),
        #
        'WinBioRegisterEventMonitor': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="HRESULT"), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Parameters": SimUnion({"Unclaimed": SimStruct({"UnitId": SimTypeInt(signed=False, label="UInt32"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_Unclaimed_e__Struct", pack=False, align=None), "UnclaimedIdentify": SimStruct({"UnitId": SimTypeInt(signed=False, label="UInt32"), "Identity": SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), "SubFactor": SimTypeChar(label="Byte"), "RejectDetail": SimTypeInt(signed=False, label="UInt32")}, name="_UnclaimedIdentify_e__Struct", pack=False, align=None), "Error": SimStruct({"ErrorCode": SimTypeBottom(label="HRESULT")}, name="_Error_e__Struct", pack=False, align=None)}, name="<anon>", label="None")}, name="WINBIO_EVENT", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["EventCallbackContext", "OperationStatus", "Event"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "EventMask", "EventCallback", "EventCallbackContext"]),
        #
        'WinBioUnregisterEventMonitor': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle"]),
        #
        'WinBioMonitorPresence': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "UnitId"]),
        #
        'WinBioCaptureSample': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeChar(label="Byte"), SimTypeChar(label="Byte"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"HeaderBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "StandardDataBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "VendorDataBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "SignatureBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None)}, name="WINBIO_BIR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "Purpose", "Flags", "UnitId", "Sample", "SampleSize", "RejectDetail"]),
        #
        'WinBioCaptureSampleWithCallback': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeChar(label="Byte"), SimTypeChar(label="Byte"), SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="HRESULT"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"HeaderBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "StandardDataBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "VendorDataBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None), "SignatureBlock": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Offset": SimTypeInt(signed=False, label="UInt32")}, name="WINBIO_BIR_DATA", pack=False, align=None)}, name="WINBIO_BIR", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeBottom(label="Void"), arg_names=["CaptureCallbackContext", "OperationStatus", "UnitId", "Sample", "SampleSize", "RejectDetail"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "Purpose", "Flags", "CaptureCallback", "CaptureCallbackContext"]),
        #
        'WinBioDeleteTemplate': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "UnitId", "Identity", "SubFactor"]),
        #
        'WinBioLockUnit': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "UnitId"]),
        #
        'WinBioUnlockUnit': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "UnitId"]),
        #
        'WinBioControlUnit': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="WINBIO_COMPONENT"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "UnitId", "Component", "ControlCode", "SendBuffer", "SendBufferSize", "ReceiveBuffer", "ReceiveBufferSize", "ReceiveDataSize", "OperationStatus"]),
        #
        'WinBioControlUnitPrivileged': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="WINBIO_COMPONENT"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "UnitId", "Component", "ControlCode", "SendBuffer", "SendBufferSize", "ReceiveBuffer", "ReceiveBufferSize", "ReceiveDataSize", "OperationStatus"]),
        #
        'WinBioGetProperty': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "PropertyType", "PropertyId", "UnitId", "Identity", "SubFactor", "PropertyBuffer", "PropertyBufferSize"]),
        #
        'WinBioSetProperty': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle", "PropertyType", "PropertyId", "UnitId", "Identity", "SubFactor", "PropertyBuffer", "PropertyBufferSize"]),
        #
        'WinBioFree': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Address"]),
        #
        'WinBioSetCredential': SimTypeFunction([SimTypeInt(signed=False, label="WINBIO_CREDENTIAL_TYPE"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="WINBIO_CREDENTIAL_FORMAT")], SimTypeInt(signed=True, label="Int32"), arg_names=["Type", "Credential", "CredentialSize", "Format"]),
        #
        'WinBioRemoveCredential': SimTypeFunction([SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), SimTypeInt(signed=False, label="WINBIO_CREDENTIAL_TYPE")], SimTypeInt(signed=True, label="Int32"), arg_names=["Identity", "Type"]),
        #
        'WinBioRemoveAllCredentials': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'WinBioRemoveAllDomainCredentials': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'WinBioGetCredentialState': SimTypeFunction([SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), SimTypeInt(signed=False, label="WINBIO_CREDENTIAL_TYPE"), SimTypePointer(SimTypeInt(signed=False, label="WINBIO_CREDENTIAL_STATE"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Identity", "Type", "CredentialState"]),
        #
        'WinBioLogonIdentifiedUser': SimTypeFunction([SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["SessionHandle"]),
        #
        'WinBioGetEnrolledFactors': SimTypeFunction([SimTypePointer(SimStruct({"Type": SimTypeInt(signed=False, label="UInt32"), "Value": SimUnion({"Null": SimTypeInt(signed=False, label="UInt32"), "Wildcard": SimTypeInt(signed=False, label="UInt32"), "TemplateGuid": SimTypeBottom(label="Guid"), "AccountSid": SimStruct({"Size": SimTypeInt(signed=False, label="UInt32"), "Data": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 68)}, name="_AccountSid_e__Struct", pack=False, align=None), "SecureId": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32)}, name="<anon>", label="None")}, name="WINBIO_IDENTITY", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["AccountOwner", "EnrolledFactors"]),
        #
        'WinBioGetEnabledSetting': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="WINBIO_SETTING_SOURCE"), offset=0)], SimTypeBottom(label="Void"), arg_names=["Value", "Source"]),
        #
        'WinBioGetLogonSetting': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="WINBIO_SETTING_SOURCE"), offset=0)], SimTypeBottom(label="Void"), arg_names=["Value", "Source"]),
        #
        'WinBioGetDomainLogonSetting': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="WINBIO_SETTING_SOURCE"), offset=0)], SimTypeBottom(label="Void"), arg_names=["Value", "Source"]),
        #
        'WinBioAcquireFocus': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'WinBioReleaseFocus': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
    }

lib.set_prototypes(prototypes)



#!/usr/bin/env python3
__package__ = __package__ or "tests.storage"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ..common import bin_location

test_location = os.path.join(bin_location, "tests")


class TestRelro(unittest.TestCase):
    def _run_fauxware_relro(self, arch):
        p = angr.Project(os.path.join(test_location, arch, "fauxware"), use_sim_procedures=False)
        s = p.factory.full_init_state(add_options={angr.options.STRICT_PAGE_ACCESS})

        relro_segment = next((s for s in p.loader.main_object.segments if s.relro), None)
        if relro_segment is None:
            # No relro on this arch
            return

        assert not relro_segment.is_writable, "The RELRO segment should not be writable"

        try:
            s.memory.store(relro_segment.min_addr, b"\x42")
            assert False, "The RELRO segment should not be writable"
        except angr.errors.SimSegfaultException:
            pass

    def test_fauxware_i386(self):
        self._run_fauxware_relro("i386")

    def test_fauxware_x86_64(self):
        self._run_fauxware_relro("x86_64")

    def test_fauxware_ppc(self):
        self._run_fauxware_relro("ppc")

    def test_fauxware_armel(self):
        self._run_fauxware_relro("armel")

    def test_fauxware_mips(self):
        self._run_fauxware_relro("mips")


if __name__ == "__main__":
    unittest.main()



# pylint:disable=unused-argument
from typing import Set, Tuple, Optional

import ailment

from .sequence_walker import SequenceWalker


class JumpTargetCollector:
    """
    Collect all jump targets.
    """

    def __init__(self, node):
        self.root = node
        self.jump_targets: Set[Tuple[int, Optional[int]]] = set()

        handlers = {
            ailment.Block: self._handle_Block,
        }
        self._walker = SequenceWalker(handlers=handlers)
        self._walker.walk(self.root)

    #
    # Handlers
    #

    def _handle_Block(self, block: ailment.Block, **kwargs):
        for stmt in block.statements:
            if isinstance(stmt, ailment.Stmt.Jump):
                if isinstance(stmt.target, ailment.Expr.Const):
                    self.jump_targets.add((stmt.target.value, stmt.target_idx))
            elif isinstance(stmt, ailment.Stmt.ConditionalJump):
                if isinstance(stmt.true_target, ailment.Expr.Const):
                    self.jump_targets.add((stmt.true_target.value, None))
                if isinstance(stmt.false_target, ailment.Expr.Const):
                    self.jump_targets.add((stmt.false_target.value, None))



from typing import Iterable, Tuple, Any, Callable, Optional

from . import MemoryMixin


class MultiValueMergerMixin(MemoryMixin):
    def __init__(self, *args, element_limit=5, annotation_limit=256, top_func=None, phi_maker=None, **kwargs):
        self._element_limit = element_limit
        self._annotation_limit = annotation_limit
        self._top_func: Callable = top_func
        self._phi_maker: Optional[Callable] = phi_maker

        super().__init__(*args, **kwargs)

    def _merge_values(self, values: Iterable[Tuple[Any, Any]], merged_size: int, **kwargs):
        values_set = {v for v, _ in values}
        if self._phi_maker is not None:
            phi_var = self._phi_maker(values_set)
            if phi_var is not None:
                return {phi_var}

        # try to merge it in the traditional way
        if len(values_set) > self._element_limit:
            # strip annotations from each value and see how many raw values there are in total
            # We have to use cache_key to determine uniqueness here, because if __hash__ collides,
            # python implicitly calls __eq__ to determine if the two objects are actually the same
            # and that just results in a new AST for a BV. Python then tries to convert that AST to a bool
            # which fails with the safeguard in claripy.ast.bool.Bool.__bool__.
            stripped_values_set = {v._apply_to_annotations(lambda alist: None).cache_key for v in values_set}
            if len(stripped_values_set) > 1:
                ret_val = self._top_func(merged_size * self.state.arch.byte_width)
            else:
                # Get the AST back from the cache_key
                ret_val = next(iter(stripped_values_set)).ast
            # migrate annotations
            annotations = []
            annotations_set = set()
            for v in values_set:
                for anno in v.annotations:
                    if anno not in annotations_set:
                        annotations.append(anno)
                        annotations_set.add(anno)
            if annotations:
                ret_val = ret_val.annotate(*annotations[: self._annotation_limit])
            merged_val = {ret_val}
        else:
            merged_val = values_set
        return merged_val

    def copy(self, memo=None):
        copied = super().copy(memo)
        copied._element_limit = self._element_limit
        copied._annotation_limit = self._annotation_limit
        copied._top_func = self._top_func
        copied._phi_maker = self._phi_maker
        return copied



# pylint:disable=no-member
import logging

l = logging.getLogger(name=__name__)

from ..protos.primitives_pb2 import Edge


_CFG_JUMPKINDS = {
    None: Edge.UnknownJumpkind,
    "Ijk_Boring": Edge.Boring,
    "Ijk_Call": Edge.Call,
    "Ijk_Ret": Edge.Return,
    "Ijk_FakeRet": Edge.FakeReturn,
    "Ijk_Syscall": Edge.Syscall,
    "Ijk_Sys_syscall": Edge.Sys_syscall,
    "Ijk_Sys_int128": Edge.Sys_int128,
    "Ijk_NoDecode": Edge.NoDecode,
    "Ijk_EmWarn": Edge.EmWarn,
    "Ijk_SigFPE_IntDiv": Edge.SigFPE_IntDiv,
    "Ijk_SigTRAP": Edge.SigTRAP,
    "Ijk_SigSEGV": Edge.SigSEGV,
    "Ijk_MapFail": Edge.MapFail,
    "Ijk_NoRedir": Edge.NoRedir,
    "Ijk_ClientReq": Edge.ClientReq,
    "Ijk_Exception": Edge.Exception,
}


_PB_TO_CFG_JUMPKINDS = {}
for k, v in _CFG_JUMPKINDS.items():
    _PB_TO_CFG_JUMPKINDS[v] = k


_FUNCTION_EDGETYPES = {
    None: Edge.UnknownJumpkind,
    "transition": Edge.Boring,
    "call": Edge.Call,
    "return": Edge.Return,
    "fake_return": Edge.FakeReturn,
    "syscall": Edge.Syscall,
    "exception": Edge.Exception,
}


_PB_TO_FUNCTION_EDGETYPES = {}
for k, v in _FUNCTION_EDGETYPES.items():
    _PB_TO_FUNCTION_EDGETYPES[v] = k


def cfg_jumpkind_to_pb(jk):
    try:
        return _CFG_JUMPKINDS[jk]
    except KeyError:
        l.error("Unsupported CFG jumpkind %s in cfg_jumpkind_to_pb. Please report it on GitHub.", jk)
        return None


def func_edge_type_to_pb(jk):
    try:
        return _FUNCTION_EDGETYPES[jk]
    except KeyError:
        l.error("Unsupported function edge type %s in func_edge_type_to_pb. Please report it on GitHub.", jk)
        return None


def cfg_jumpkind_from_pb(pb):
    try:
        return _PB_TO_CFG_JUMPKINDS[pb]
    except KeyError:
        l.error("Unsupported protobuf jumpkind %s in cfg_jumpkind_from_pb. Please report it on GitHub.", pb)
        return None


def func_edge_type_from_pb(pb):
    try:
        return _PB_TO_FUNCTION_EDGETYPES[pb]
    except KeyError:
        l.error("Unsupported protobuf jumpkind %s in func_edge_type_to_pb. Please report it on GitHub.", pb)
        return None



from .regioned_memory_mixin import RegionedMemoryMixin
from .region_category_mixin import RegionCategoryMixin
from .static_find_mixin import StaticFindMixin
from .abstract_merger_mixin import AbstractMergerMixin
from .region_meta_mixin import MemoryRegionMetaMixin
from .regioned_address_concretization_mixin import RegionedAddressConcretizationMixin



import logging

import angr


l = logging.getLogger(name=__name__)

# BOOL GetLastInputInfo(
#   PLASTINPUTINFO plii
# );

# | typedef struct tagLASTINPUTINFO {
# |   UINT  cbSize;  // The size of the structure, in bytes.
# |                  // This member must be set to sizeof(LASTINPUTINFO).
# |   DWORD dwTime;
# | } LASTINPUTINFO, *PLASTINPUTINFO;


class GetLastInputInfo(angr.SimProcedure):
    cbSize = None
    dwTime = None

    def run(self, plii):  # pylint:disable=arguments-differ
        self.fill_symbolic()
        l.info("Setting symbolic memory at %s", str(plii))
        self.state.mem[plii].dword = self.cbSize
        self.state.mem[plii + 4].dword = self.dwTime

        return 1

    def fill_symbolic(self):
        self.cbSize = self.state.solver.BVS("tagLASTINPUTINFO_cbSize", 32, key=("api", "tagLASTINPUTINFO_cbSize"))
        self.dwTime = self.state.solver.BVS("tagLASTINPUTINFO_dwTime", 32, key=("api", "tagLASTINPUTINFO_dwTime"))

    def fill_concrete(self):
        self.cbSize = self.state.solver.BVV(3, 32)
        self.dwTime = self.state.solver.BVV(3, 32)



import angr


class openat(angr.SimProcedure):
    # pylint:disable=arguments-differ,unused-argument

    def run(self, dirfd, p_addr, flags, mode):
        strlen = angr.SIM_PROCEDURES["libc"]["strlen"]

        p_strlen = self.inline_call(strlen, p_addr)
        p_expr = self.state.memory.load(p_addr, p_strlen.max_null_index, endness="Iend_BE")
        path = self.state.solver.eval(p_expr, cast_to=bytes)

        # If path is absolute or dirfd is AT_FDCWD(-100), dirfd can be ignored.
        # Sometimes dirfd can be a 32 bit value in a 64 bit BV. So, instead of converting dirfd to signed integer, we
        # simply check the unsigned value.
        # TODO: Is above described way to check dirfd okay?
        dirfd_val = self.state.solver.eval(dirfd)
        if path.startswith(b"/") or dirfd_val == 0xFFFFFF9C:
            fd = self.state.posix.open(path, flags)
        else:
            # TODO: Implement support for opening path relative to directory corresponding to dirfd
            fd = None

        if fd is None:
            return -1

        return fd



from typing import TYPE_CHECKING

import archinfo

if TYPE_CHECKING:
    from angr import Project


def is_pc(project: "Project", ins_addr: int, addr: int) -> bool:
    """
    Check if the given address is program counter (PC) or not. This function is for handling the case on some bizarre
    architectures where PC is always the currently executed instruction address plus a constant value.

    :param project:     An angr Project instance.
    :param ins_addr:    The address of an instruction. We calculate PC using this instruction address.
    :param addr:        The address to check against.
    :return:            True if the given instruction address is the PC, False otherwise.
    """
    if archinfo.arch_arm.is_arm_arch(project.arch):
        if ins_addr & 1 == 1:
            # thumb mode
            ins_addr = ins_addr - 1
            return addr == ins_addr + 4
        else:
            # arm mode
            return addr == ins_addr + 8
    return ins_addr == addr


def is_in_readonly_section(project: "Project", addr: int) -> bool:
    """
    Check if the specified address is inside a read-only section.

    :param project:     An angr Project instance.
    :param addr:        The address to check.
    :return:            True if the given address belongs to a read-only section, False otherwise.
    """
    sec = project.loader.find_section_containing(addr)
    if sec is not None:
        return not sec.is_writable
    return False


def is_in_readonly_segment(project: "Project", addr: int) -> bool:
    """
    Check if the specified address is inside a read-only segment.

    :param project:     An angr Project instance.
    :param addr:        The address to check.
    :return:            True if the given address belongs to a read-only segment, False otherwise.
    """
    seg = project.loader.find_segment_containing(addr)
    if seg is not None:
        return not seg.is_writable
    return False



from ...errors import SimSolverError


def concretize(x, solver, sym_handler):
    """
    For now a lot of naive concretization is done when handling heap metadata to keep things manageable. This idiom
    showed up a lot as a result, so to reduce code repetition this function uses a callback to handle the one or two
    operations that varied across invocations.

    :param x: the item to be concretized
    :param solver: the solver to evaluate the item with
    :param sym_handler: the handler to be used when the item may take on more than one value
    :returns: a concrete value for the item
    """
    if solver.symbolic(x):
        try:
            return solver.eval_one(x)
        except SimSolverError:
            return sym_handler(x)
    else:
        return solver.eval(x)



import angr
import itertools


malloc_mem_counter = itertools.count()


class malloc(angr.SimProcedure):
    # pylint:disable=arguments-differ
    def run(self, sim_size):
        return self.state.heap._malloc(sim_size)



#!/usr/bin/env python3
import unittest

from angr import SimState, SimHeapPTMalloc


# TODO: Make these tests more architecture-independent (note dependencies of some behavior on chunk metadata size)
class TestPtmalloc(unittest.TestCase):
    def chunk_iterators_are_same(self, iterator1, iterator2):
        for ck in iterator1:
            ck2 = next(iterator2)
            if ck.base != ck2.base:
                return False
            if ck.is_free() != ck2.is_free():
                return False
        try:
            next(iterator2)
        except StopIteration:
            return True
        return False

    def same_heap_states(self, state1, state2):
        return self.chunk_iterators_are_same(state1.heap.chunks(), state2.heap.chunks())

    def max_sym_var_val(self, state):
        return state.libc.max_variable_size

    def _run_malloc_maximizes_sym_arg(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        sc = s.copy()
        x = s.solver.BVS("x", 32)
        s.solver.add(x.UGE(0))
        s.solver.add(x.ULE(self.max_sym_var_val(s)))
        s.heap.malloc(x)
        sc.heap.malloc(self.max_sym_var_val(sc))
        assert self.same_heap_states(s, sc)

    def test_malloc_maximizes_sym_arg_X86(self):
        self._run_free_maximizes_sym_arg("X86")

    def test_malloc_maximizes_sym_arg_AMD64(self):
        self._run_free_maximizes_sym_arg("AMD64")

    def _run_free_maximizes_sym_arg(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        p = s.heap.malloc(50)
        sc = s.copy()
        x = s.solver.BVS("x", 32)
        s.solver.add(x.UGE(0))
        s.solver.add(x.ULE(p))
        s.heap.free(x)
        sc.heap.free(p)
        assert self.same_heap_states(s, sc)

    def test_free_maximizes_sym_arg_X86(self):
        self._run_free_maximizes_sym_arg("X86")

    def test_free_maximizes_sym_arg_AMD64(self):
        self._run_free_maximizes_sym_arg("AMD64")

    def _run_calloc_maximizes_sym_arg(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        sc = s.copy()
        x = s.solver.BVS("x", 32)
        s.solver.add(x.UGE(0))
        s.solver.add(x.ULE(20))
        y = s.solver.BVS("y", 32)
        s.solver.add(y.UGE(0))
        s.solver.add(y.ULE(6))
        s.heap.calloc(x, y)
        sc.heap.calloc(20, 6)
        assert self.same_heap_states(s, sc)

    def test_calloc_maximizes_sym_arg_X86(self):
        self._run_calloc_maximizes_sym_arg("X86")

    def test_calloc_maximizes_sym_arg_AMD64(self):
        self._run_calloc_maximizes_sym_arg("AMD64")

    def _run_realloc_maximizes_sym_arg(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        p = s.heap.malloc(50)
        sc = s.copy()
        x = s.solver.BVS("x", 32)
        s.solver.add(x.UGE(0))
        s.solver.add(x.ULE(p))
        y = s.solver.BVS("y", 32)
        s.solver.add(y.UGE(0))
        s.solver.add(y.ULE(self.max_sym_var_val(s)))
        s.heap.realloc(x, y)
        sc.heap.realloc(p, self.max_sym_var_val(sc))
        assert self.same_heap_states(s, sc)

    def test_realloc_maximizes_sym_arg_X86(self):
        self._run_realloc_maximizes_sym_arg("X86")

    def test_realloc_maximizes_sym_arg_AMD64(self):
        self._run_realloc_maximizes_sym_arg("AMD64")

    def _run_malloc_no_space_returns_null(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        sc = s.copy()
        p1 = s.heap.malloc(0x2000)
        assert p1 == 0
        assert self.same_heap_states(s, sc)

    def test_malloc_no_space_returns_null_X86(self):
        self._run_malloc_no_space_returns_null("X86")

    def test_malloc_no_space_returns_null_AMD64(self):
        self._run_malloc_no_space_returns_null("AMD64")

    def _run_calloc_no_space_returns_null(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        sc = s.copy()
        p1 = s.heap.calloc(0x500, 4)
        assert p1 == 0
        assert self.same_heap_states(s, sc)

    def test_calloc_no_space_returns_null_X86(self):
        self._run_calloc_no_space_returns_null("X86")

    def test_calloc_no_space_returns_null_AMD64(self):
        self._run_calloc_no_space_returns_null("AMD64")

    def _run_realloc_no_space_returns_null(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        p1 = s.heap.malloc(20)
        sc = s.copy()
        p2 = s.heap.realloc(p1, 0x2000)
        assert p2 == 0
        assert self.same_heap_states(s, sc)

    def test_realloc_no_space_returns_null_X86(self):
        self._run_realloc_no_space_returns_null("X86")

    def test_realloc_no_space_returns_null_AMD64(self):
        self._run_realloc_no_space_returns_null("AMD64")

    def _run_first_fit_and_free_malloced_makes_available(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        s.heap.malloc(20)
        p1 = s.heap.malloc(50)
        s.heap.free(p1)
        p2 = s.heap.malloc(30)
        assert p1 == p2

    def test_first_fit_and_free_malloced_makes_available_X86(self):
        self._run_first_fit_and_free_malloced_makes_available("X86")

    def test_first_fit_and_free_malloced_makes_available_AMD64(self):
        self._run_first_fit_and_free_malloced_makes_available("AMD64")

    def _run_free_calloced_makes_available(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        s.heap.calloc(20, 5)
        p1 = s.heap.calloc(30, 4)
        s.heap.free(p1)
        p2 = s.heap.calloc(15, 8)
        assert p1 == p2

    def test_free_calloced_makes_available_X86(self):
        self._run_free_calloced_makes_available("X86")

    def test_free_calloced_makes_available_AMD64(self):
        self._run_free_calloced_makes_available("AMD64")

    def _run_realloc_moves_and_frees(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        s.heap.malloc(20)
        p1 = s.heap.malloc(60)
        s.heap.malloc(200)
        p2 = s.heap.realloc(p1, 300)
        p3 = s.heap.malloc(30)
        assert p1 == p3
        assert p1 < p2

    def test_realloc_moves_and_frees_X86(self):
        self._run_realloc_moves_and_frees("X86")

    def test_realloc_moves_and_frees_AMD64(self):
        self._run_realloc_moves_and_frees("AMD64")

    def _run_realloc_near_same_size(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        s.heap.malloc(20)
        p1 = s.heap.malloc(61)
        s.heap.malloc(80)
        sc = s.copy()
        p2 = s.heap.realloc(p1, 62)
        assert p1 == p2
        assert self.same_heap_states(s, sc)

    def test_realloc_near_same_size_X86(self):
        self._run_realloc_near_same_size("X86")

    def test_realloc_near_same_size_AMD64(self):
        self._run_realloc_near_same_size("AMD64")

    def _run_needs_space_for_metadata(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        sc = s.copy()
        p1 = s.heap.malloc(0x1000)
        assert p1 == 0
        assert self.same_heap_states(s, sc)

    def test_needs_space_for_metadata_X86(self):
        self._run_needs_space_for_metadata("X86")

    def test_needs_space_for_metadata_AMD64(self):
        self._run_needs_space_for_metadata("AMD64")

    def _run_unusable_amount_returns_null(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        s.heap.malloc(0x1000 - 4 * s.heap._chunk_size_t_size)
        sc = s.copy()
        p = s.heap.malloc(1)
        assert p == 0
        assert self.same_heap_states(s, sc)

    def test_unusable_amount_returns_null_X86(self):
        self._run_unusable_amount_returns_null("X86")

    def test_unusable_amount_returns_null_AMD64(self):
        self._run_unusable_amount_returns_null("AMD64")

    def _run_free_null_preserves_state(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        s.heap.malloc(30)
        p = s.heap.malloc(40)
        s.heap.malloc(50)
        s.heap.free(p)
        s2 = s.copy()
        s2.heap.free(0)
        assert self.same_heap_states(s, s2)

    def test_free_null_preserves_state_X86(self):
        self._run_free_null_preserves_state("X86")

    def test_free_null_preserves_state_AMD64(self):
        self._run_free_null_preserves_state("AMD64")

    def _run_skips_chunks_too_small(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        s.heap.malloc(30)
        p = s.heap.malloc(50)
        s.heap.malloc(40)
        s.heap.free(p)
        p2 = s.heap.calloc(20, 5)
        assert p < p2

    def test_skips_chunks_too_small_X86(self):
        self._run_skips_chunks_too_small("X86")

    def test_skips_chunks_too_small_AMD64(self):
        self._run_skips_chunks_too_small("AMD64")

    def _run_calloc_multiplies(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        s.heap.malloc(30)
        sc = s.copy()
        s.heap.malloc(100)
        sc.heap.calloc(4, 25)
        assert self.same_heap_states(s, sc)

    def test_calloc_multiplies_X86(self):
        self._run_calloc_multiplies("X86")

    def test_calloc_multiplies_AMD64(self):
        self._run_calloc_clears("AMD64")

    def _run_calloc_clears(self, arch):
        s = SimState(arch=arch, plugins={"heap": SimHeapPTMalloc(heap_base=0xD0000000, heap_size=0x1000)})
        s.memory.store(0xD0000000 + 2 * s.heap._chunk_size_t_size, s.solver.BVV(-1, 100 * 8))
        sc = s.copy()
        p1 = s.heap.calloc(6, 5)
        p2 = sc.heap.malloc(30)
        v1 = s.memory.load(p1, 30)
        v2 = sc.memory.load(p2, 30)
        assert s.solver.is_true(v1 == 0)
        assert sc.solver.is_true(v2 == -1)

    def test_calloc_clears_X86(self):
        self._run_calloc_clears("X86")

    def test_calloc_clears_AMD64(self):
        self._run_calloc_clears("AMD64")


if __name__ == "__main__":
    unittest.main()



import angr

from cle.backends.externs.simdata.io_file import io_file_data_for_arch


class ungetc(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, c, file_ptr):
        # TODO handle errors
        # TODO THIS DOESN'T WORK IN ANYTHING BUT THE TYPICAL CASE
        fd_offset = io_file_data_for_arch(self.state.arch)["fd"]
        fileno = self.state.mem[file_ptr + fd_offset :].int.concrete
        if hasattr(self.state.posix.fd[fileno], "_read_pos"):
            self.state.posix.fd[fileno]._read_pos -= 1
        else:
            self.state.posix.fd[fileno]._pos -= 1

        return c & 0xFF



#!/usr/bin/env python3
__package__ = __package__ or "tests.analyses"  # pylint:disable=redefined-builtin

import logging
import os
import unittest

import archinfo

import angr
from angr.calling_conventions import (
    SimStackArg,
    SimRegArg,
    SimCCCdecl,
    SimCCSystemVAMD64,
)
from angr.sim_type import SimTypeFunction, SimTypeInt, SimTypeLongLong, SimTypeBottom

from ..common import bin_location, requires_binaries_private


test_location = os.path.join(bin_location, "tests")


# pylint: disable=missing-class-docstring
# pylint: disable=no-self-use
class TestCallingConventionAnalysis(unittest.TestCase):
    def _run_fauxware(self, arch, function_and_cc_list):
        binary_path = os.path.join(test_location, arch, "fauxware")
        fauxware = angr.Project(binary_path, auto_load_libs=False)

        cfg = fauxware.analyses.CFG()

        for func_name, expected_cc in function_and_cc_list:
            authenticate = cfg.functions[func_name]
            _ = fauxware.analyses.VariableRecoveryFast(authenticate)

            cc_analysis = fauxware.analyses.CallingConvention(authenticate, cfg=cfg, analyze_callsites=True)
            cc = cc_analysis.cc

            assert cc == expected_cc

    def _run_cgc(self, binary_name):
        pass
        # binary_path = os.path.join(bin_location, '..', 'binaries-private', 'cgc_qualifier_event', 'cgc', binary_name)
        # project = angr.Project(binary_path, auto_load_libs=False)
        #
        # categorization = project.analyses.FunctionCategorizationAnalysis()

        # tag_manager = categorization.function_tag_manager
        # print "INPUT:", map(hex, tag_manager.input_functions())
        # print "OUTPUT:", map(hex, tag_manager.output_functions())

    def test_fauxware_i386(self):
        self._run_fauxware("i386", [("authenticate", SimCCCdecl(archinfo.arch_from_id("i386")))])

    def test_fauxware_x86_64(self):
        amd64 = archinfo.arch_from_id("amd64")
        self._run_fauxware(
            "x86_64",
            [
                (
                    "authenticate",
                    SimCCSystemVAMD64(
                        amd64,
                    ),
                ),
            ],
        )

    @requires_binaries_private
    def test_cgc_binary1(self):
        self._run_cgc("002ba801_01")

    @requires_binaries_private
    def test_cgc_binary2(self):
        self._run_cgc("01cf6c01_01")

    #
    # Full-binary calling convention analysis
    #

    def check_arg(self, arg, expected_str):
        if isinstance(arg, SimRegArg):
            arg_str = "r_%s" % (arg.reg_name)
        else:
            raise TypeError("Unsupported argument type %s." % type(arg))
        return arg_str == expected_str

    def check_args(self, func_name, args, expected_arg_strs):
        assert len(args) == len(expected_arg_strs), "Wrong number of arguments for function %s. Got %d, expect %d." % (
            func_name,
            len(args),
            len(expected_arg_strs),
        )

        for idx, (arg, expected_arg_str) in enumerate(zip(args, expected_arg_strs)):
            r = self.check_arg(arg, expected_arg_str)
            assert r, "Incorrect argument %d for function %s. Got %s, expect %s." % (
                idx,
                func_name,
                arg,
                expected_arg_str,
            )

    def _a(self, funcs, func_name):
        func = funcs[func_name]
        return func.calling_convention.arg_locs(func.prototype)

    def test_x8664_dir_gcc_O0(self):
        binary_path = os.path.join(test_location, "x86_64", "dir_gcc_-O0")
        proj = angr.Project(binary_path, auto_load_libs=False, load_debug_info=False)

        cfg = proj.analyses.CFG()  # fill in the default kb

        proj.analyses.CompleteCallingConventions(recover_variables=True)

        funcs = cfg.kb.functions

        # check args
        expected_args = {
            "c_ispunct": ["r_rdi"],
            "file_failure": ["r_rdi", "r_rsi", "r_rdx"],
            "to_uchar": ["r_rdi"],
            "dot_or_dotdot": ["r_rdi"],
            "emit_mandatory_arg_note": [],
            "emit_size_note": [],
            "emit_ancillary_info": ["r_rdi"],
            "emit_try_help": [],
            "dev_ino_push": ["r_rdi", "r_rsi"],
            "main": ["r_rdi", "r_rsi"],
            "queue_directory": ["r_rdi", "r_rsi", "r_rdx"],
        }

        for func_name, args in expected_args.items():
            self.check_args(func_name, self._a(funcs, func_name), args)

    def test_armel_fauxware(self):
        binary_path = os.path.join(test_location, "armel", "fauxware")
        proj = angr.Project(binary_path, auto_load_libs=False, load_debug_info=False)

        cfg = proj.analyses.CFG()  # fill in the default kb

        proj.analyses.CompleteCallingConventions(recover_variables=True)

        funcs = cfg.kb.functions

        # check args
        expected_args = {
            "main": ["r_r0", "r_r1"],
            "accepted": ["r_r0", "r_r1", "r_r2", "r_r3"],
            "rejected": [],
            "authenticate": ["r_r0", "r_r1", "r_r2"],  # TECHNICALLY WRONG but what are you gonna do about it
            # details: open(3) can take either 2 or 3 args. we use the 2 arg version but we have the 3 arg version
            # hardcoded in angr. the third arg is still "live" from the function start.
        }

        for func_name, args in expected_args.items():
            self.check_args(func_name, self._a(funcs, func_name), args)

    def test_x8664_void(self):
        binary_path = os.path.join(test_location, "x86_64", "types", "void")
        proj = angr.Project(binary_path, auto_load_libs=False, load_debug_info=False)

        cfg = proj.analyses.CFG()

        proj.analyses.CompleteCallingConventions(recover_variables=True, cfg=cfg.model, analyze_callsites=True)

        funcs = cfg.kb.functions

        groundtruth = {
            "func_1": None,
            "func_2": None,
            "func_3": "rax",
            "func_4": None,
            "func_5": None,
            "func_6": "rax",
        }

        for func in funcs.values():
            if func.is_simprocedure or func.alignment:
                continue
            if func.calling_convention is None:
                continue
            if func.name in groundtruth:
                r = groundtruth[func.name]
                if r is None:
                    assert isinstance(func.prototype.returnty, SimTypeBottom)
                else:
                    ret_val = func.calling_convention.return_val(func.prototype.returnty)
                    assert isinstance(ret_val, SimRegArg)
                    assert ret_val.reg_name == r

    def test_x86_saved_regs(self):
        # Calling convention analysis should be able to determine calling convention of functions with registers
        # saved on the stack.
        binary_path = os.path.join(test_location, "cgc", "NRFIN_00036")
        proj = angr.Project(binary_path, auto_load_libs=False)

        cfg = proj.analyses.CFG()
        func = cfg.functions[0x80494F0]  # int2str

        proj.analyses.VariableRecoveryFast(func)
        cca = proj.analyses.CallingConvention(func)
        cc = cca.cc
        prototype = cca.prototype

        assert cc is not None, (
            "Calling convention analysis failed to determine the calling convention of function " "0x80494f0."
        )
        assert isinstance(cc, SimCCCdecl)
        assert len(prototype.args) == 3
        arg_locs = cc.arg_locs(prototype)
        assert arg_locs[0] == SimStackArg(4, 4)
        assert arg_locs[1] == SimStackArg(8, 4)
        assert arg_locs[2] == SimStackArg(12, 4)

        func_exit = cfg.functions[0x804A1A9]  # exit

        proj.analyses.VariableRecoveryFast(func_exit)
        cca = proj.analyses.CallingConvention(func_exit)
        cc = cca.cc
        prototype = cca.prototype

        assert func_exit.returning is False
        assert cc is not None, (
            "Calling convention analysis failed to determine the calling convention of function " "0x804a1a9."
        )
        assert isinstance(cc, SimCCCdecl)
        assert len(prototype.args) == 1
        assert cc.arg_locs(prototype)[0] == SimStackArg(4, 4)

    def test_callsite_inference_amd64(self):
        # Calling convention analysis should be able to determine calling convention of a library function by
        # analyzing its callsites.
        binary_path = os.path.join(test_location, "x86_64", "decompiler", "morton")
        proj = angr.Project(binary_path, auto_load_libs=False)
        cfg = proj.analyses.CFG(data_references=True, normalize=True)

        func = cfg.functions.function(name="mosquitto_publish", plt=True)
        proj.analyses.VariableRecoveryFast(func)
        cca = proj.analyses.CallingConvention(func, analyze_callsites=True)
        assert len(cca.prototype.args) == 6

    def test_x64_return_value_used(self):
        binary_path = os.path.join(test_location, "x86_64", "cwebp-0.3.1-feh-original")
        proj = angr.Project(binary_path, auto_load_libs=False)
        cfg = proj.analyses.CFGFast(normalize=True, data_references=True, force_complete_scan=False)
        func = proj.kb.functions.get_by_addr(0x4046E0)
        proj.analyses.VariableRecoveryFast(func)
        cca = proj.analyses.CallingConvention(func=func, cfg=cfg, analyze_callsites=True)

        assert cca.prototype is not None
        assert cca.prototype.returnty is not None

    def test_armhf_thumb_movcc(self):
        binary_path = os.path.join(test_location, "armhf", "amp_challenge_07.gcc")
        proj = angr.Project(binary_path, auto_load_libs=False)
        _ = proj.analyses.CFGFast(normalize=True, data_references=True, regions=[(0xFEC94, 0xFEF60)])
        f = proj.kb.functions[0xFEC95]
        proj.analyses.VariableRecoveryFast(f)
        cca = proj.analyses.CallingConvention(f)

        assert cca.prototype is not None
        assert cca.cc is not None
        assert isinstance(cca.prototype, SimTypeFunction)
        assert len(cca.prototype.args) == 2

    def manual_test_workers(self):
        binary_path = os.path.join(test_location, "x86_64", "1after909")
        proj = angr.Project(binary_path, auto_load_libs=False, load_debug_info=False)

        cfg = proj.analyses.CFG(normalize=True)  # fill in the default kb

        _ = proj.analyses.CompleteCallingConventions(
            cfg=cfg.model, recover_variables=True, workers=4, show_progressbar=True
        )

        for func in cfg.functions.values():
            assert func.is_prototype_guessed is True

    def test_tail_calls(self):
        for opt_level in (1, 2):
            binary_path = os.path.join(test_location, "x86_64", "tailcall-O%d" % opt_level)
            proj = angr.Project(binary_path, auto_load_libs=False)

            proj.analyses.CFG(normalize=True)
            proj.analyses.CompleteCallingConventions(recover_variables=True)

            for func in ["target", "direct", "plt"]:
                # expected prototype: (int) -> long long
                # technically should be (int) -> int, but the compiler loads all 64 bits and then truncates
                proto = proj.kb.functions[func].prototype
                assert len(proto.args) == 1
                assert isinstance(proto.args[0], SimTypeInt)
                assert isinstance(proto.returnty, SimTypeLongLong)

    def test_ls_gcc_O0_timespec_cmp(self):
        binary_path = os.path.join(test_location, "x86_64", "decompiler", "ls_gcc_O0")
        proj = angr.Project(binary_path, auto_load_libs=False)

        proj.analyses.CFG(normalize=True)
        proj.analyses.VariableRecoveryFast(proj.kb.functions["timespec_cmp"])
        cca = proj.analyses.CallingConvention(proj.kb.functions["timespec_cmp"])

        assert len(cca.prototype.args) == 4

    def test_run_multiple_times(self):
        binary_path = os.path.join(test_location, "x86_64", "fauxware")
        proj = angr.Project(binary_path, auto_load_libs=False)

        proj.analyses.CFG(normalize=True)
        proj.analyses.CompleteCallingConventions(recover_variables=True)

        expected_prototype = proj.kb.functions["main"].prototype
        proj.analyses.CompleteCallingConventions(recover_variables=True)
        assert proj.kb.functions["main"].prototype == expected_prototype

        proj.analyses.CFG(normalize=True)
        proj.analyses.CompleteCallingConventions(recover_variables=True)
        assert proj.kb.functions["main"].prototype == expected_prototype

    def test_test_three_arguments(self):
        binary_path = os.path.join(test_location, "x86_64", "test.o")
        proj = angr.Project(binary_path, auto_load_libs=False)

        cfg = proj.analyses.CFG(normalize=True)
        # the node 0x401226 must be in its own function
        assert cfg.model.get_any_node(0x401226).function_address == 0x401226

        proj.analyses.CompleteCallingConventions(recover_variables=True)

        assert proj.kb.functions["test_syntax_error"].prototype.variadic is True
        assert len(proj.kb.functions["expr"].prototype.args) == 0

    def test_windows_partial_input_variable_overwrite(self):
        binary_path = os.path.join(test_location, "x86_64", "netfilter_b64.sys")
        proj = angr.Project(binary_path, auto_load_libs=False)

        cfg = proj.analyses.CFG(normalize=True)
        proj.analyses.VariableRecoveryFast(proj.kb.functions[0x140001A90])
        cc = proj.analyses.CallingConvention(cfg.kb.functions[0x140001A90])
        assert cc.cc is not None
        assert cc.prototype is not None
        print(cc.prototype.args)
        assert len(cc.prototype.args) == 3


if __name__ == "__main__":
    # logging.getLogger("angr.analyses.variable_recovery.variable_recovery_fast").setLevel(logging.DEBUG)
    logging.getLogger("angr.analyses.calling_convention").setLevel(logging.INFO)
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long
__package__ = __package__ or "tests.analyses"  # pylint:disable=redefined-builtin

import os
import unittest

import angr

from ..common import bin_location

test_location = os.path.join(bin_location, "tests")


class TestProximityGraph(unittest.TestCase):
    def test_fauxware(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        proj = angr.Project(bin_path, auto_load_libs=False)

        cfg = proj.analyses.CFG(data_references=True, cross_references=True, normalize=True)
        func = cfg.kb.functions["main"]

        proj.analyses.Proximity(func, cfg.model, cfg.kb.xrefs)

        # once we have decompiled code, things are different...
        dec = proj.analyses.Decompiler(func, cfg=cfg.model)
        proj.analyses.Proximity(func, cfg.model, cfg.kb.xrefs, decompilation=dec)


if __name__ == "__main__":
    unittest.main()



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("d3dcompiler_47.dll")
prototypes = \
    {
        #
        'D3DDisassemble11Trace': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeBottom(label="ID3D11ShaderTrace"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "pTrace", "StartStep", "NumSteps", "Flags", "ppDisassembly"]),
        #
        'D3DReadFileToBlob': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pFileName", "ppContents"]),
        #
        'D3DWriteBlobToFile': SimTypeFunction([SimTypeBottom(label="ID3DBlob"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pBlob", "pFileName", "bOverwrite"]),
        #
        'D3DCompile': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"Name": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Definition": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="D3D_SHADER_MACRO", pack=False, align=None), offset=0), SimTypeBottom(label="ID3DInclude"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "pSourceName", "pDefines", "pInclude", "pEntrypoint", "pTarget", "Flags1", "Flags2", "ppCode", "ppErrorMsgs"]),
        #
        'D3DCompile2': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"Name": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Definition": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="D3D_SHADER_MACRO", pack=False, align=None), offset=0), SimTypeBottom(label="ID3DInclude"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "pSourceName", "pDefines", "pInclude", "pEntrypoint", "pTarget", "Flags1", "Flags2", "SecondaryDataFlags", "pSecondaryData", "SecondaryDataSize", "ppCode", "ppErrorMsgs"]),
        #
        'D3DCompileFromFile': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"Name": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Definition": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="D3D_SHADER_MACRO", pack=False, align=None), offset=0), SimTypeBottom(label="ID3DInclude"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pFileName", "pDefines", "pInclude", "pEntrypoint", "pTarget", "Flags1", "Flags2", "ppCode", "ppErrorMsgs"]),
        #
        'D3DPreprocess': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"Name": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Definition": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="D3D_SHADER_MACRO", pack=False, align=None), offset=0), SimTypeBottom(label="ID3DInclude"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "pSourceName", "pDefines", "pInclude", "ppCodeText", "ppErrorMsgs"]),
        #
        'D3DGetDebugInfo': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "ppDebugInfo"]),
        #
        'D3DReflect': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "pInterface", "ppReflector"]),
        #
        'D3DReflectLibrary': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "riid", "ppReflector"]),
        #
        'D3DDisassemble': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "Flags", "szComments", "ppDisassembly"]),
        #
        'D3DDisassembleRegion': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "Flags", "szComments", "StartByteOffset", "NumInsts", "pFinishByteOffset", "ppDisassembly"]),
        #
        'D3DCreateLinker': SimTypeFunction([SimTypePointer(SimTypeBottom(label="ID3D11Linker"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppLinker"]),
        #
        'D3DLoadModule': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3D11Module"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "cbSrcDataSize", "ppModule"]),
        #
        'D3DCreateFunctionLinkingGraph': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3D11FunctionLinkingGraph"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["uFlags", "ppFunctionLinkingGraph"]),
        #
        'D3DGetTraceInstructionOffsets': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "Flags", "StartInstIndex", "NumInsts", "pOffsets", "pTotalInsts"]),
        #
        'D3DGetInputSignatureBlob': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "ppSignatureBlob"]),
        #
        'D3DGetOutputSignatureBlob': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "ppSignatureBlob"]),
        #
        'D3DGetInputAndOutputSignatureBlob': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "ppSignatureBlob"]),
        #
        'D3DStripShader': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pShaderBytecode", "BytecodeLength", "uStripFlags", "ppStrippedBlob"]),
        #
        'D3DGetBlobPart': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="D3D_BLOB_PART"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "Part", "Flags", "ppPart"]),
        #
        'D3DSetBlobPart': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="D3D_BLOB_PART"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "Part", "Flags", "pPart", "PartSize", "ppNewShader"]),
        #
        'D3DCreateBlob': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["Size", "ppBlob"]),
        #
        'D3DCompressShaders': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"pBytecode": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "BytecodeLength": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0)}, name="D3D_SHADER_DATA", pack=False, align=None), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["uNumShaders", "pShaderData", "uFlags", "ppCompressedData"]),
        #
        'D3DDecompressShaders': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pSrcData", "SrcDataSize", "uNumShaders", "uStartIndex", "pIndices", "uFlags", "ppShaders", "pTotalShaders"]),
        #
        'D3DDisassemble10Effect': SimTypeFunction([SimTypeBottom(label="ID3D10Effect"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="ID3DBlob"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pEffect", "Flags", "ppDisassembly"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=unused-argument,arguments-differ

from ..sequence_walker import SequenceWalker
from ..structuring.structurer_nodes import (
    SequenceNode,
    CodeNode,
    MultiNode,
    LoopNode,
    ConditionNode,
    EmptyBlockNotice,
    CascadingConditionNode,
)
from ..condition_processor import ConditionProcessor
from ..utils import insert_node, is_statement_terminating


class IfElseFlattener(SequenceWalker):
    """
    Remove unnecessary else branches and make the else node a direct successor of the previous If node if the If node
    always returns.
    """

    def __init__(self, node, functions):
        handlers = {
            SequenceNode: self._handle_Sequence,
            CodeNode: self._handle_Code,
            MultiNode: self._handle_MultiNode,
            LoopNode: self._handle_Loop,
            ConditionNode: self._handle_Condition,
            CascadingConditionNode: self._handle_CascadingCondition,
        }

        super().__init__(handlers)
        self.functions = functions
        self.walk(node)

    def _handle_Condition(self, node: ConditionNode, parent=None, index=None, **kwargs):
        if node.true_node is not None:
            self._handle(node.true_node, parent=node, index=0)
        if node.false_node is not None:
            self._handle(node.false_node, parent=node, index=1)

        if node.true_node is not None and node.false_node is not None:
            try:
                last_stmts = ConditionProcessor.get_last_statements(node.true_node)
            except EmptyBlockNotice:
                last_stmts = None
            if (
                last_stmts is not None
                and None not in last_stmts
                and all(is_statement_terminating(stmt, self.functions) for stmt in last_stmts)
            ):
                # all end points in the true node are returning

                # remove the else node and make it a new node following node
                else_node = node.false_node
                node.false_node = None
                insert_node(parent, "after", else_node, index, **kwargs)

    def _handle_CascadingCondition(self, node: CascadingConditionNode, parent=None, index=None, **kwargs):
        super()._handle_CascadingCondition(node, parent=parent, index=index, **kwargs)

        if node.else_node is not None:
            last_stmts = []
            for _, subnode in node.condition_and_nodes:
                try:
                    last_stmts.extend(ConditionProcessor.get_last_statements(subnode))
                except EmptyBlockNotice:
                    last_stmts.append(None)

            if (
                last_stmts
                and None not in last_stmts
                and all(is_statement_terminating(stmt, self.functions) for stmt in last_stmts)
            ):
                # all end points in the true node are returning

                # remove the else node and make it a new node following node
                else_node = node.else_node
                node.else_node = None
                insert_node(parent, "after", else_node, index, **kwargs)
